
import uuid
from autoboot.annotation import static_property

class WebProperties:
  
  @static_property("autoboot.web.scan_controller_packages")
  def scan_controller_packages() -> list[str]:
    pass
  
  @static_property("autoboot.web.security.cors.cross_origin")
  def cross_origin() -> bool:
    return False

  @static_property("autoboot.web.security.cors.allow_origins")
  def allow_origins() -> list[str]:
    return ["*"]

  @static_property("autoboot.web.security.cors.allow_methods")
  def allow_methods() -> list[str]:
    return ["*"]

  @static_property("autoboot.web.security.cors.allow_headers")
  def allow_headers() -> list[str]:
    return ["*"]

  @static_property("autoboot.web.security.cors.allow_credentials")
  def allow_credentials() -> bool:
    return True
  
  @static_property("autoboot.web.security.cors.max_age")
  def max_age() -> int:
    return 3600

  @static_property("autoboot.web.security.csrf.enable")
  def csrf_enable() -> bool:
    return False
  
  @static_property("autoboot.web.security.csrf.secret")
  def csrf_secret() -> str:
    return str(uuid.uuid1())

  @static_property("autoboot.web.security.csrf.cookie_name")
  def csrf_cookie_name() -> str:
    return "CSRF_Token"
  
  @static_property("autoboot.web.security.csrf.csrf_cookie_domain")
  def csrf_cookie_domain() -> str:
    return None

  @static_property("autoboot.web.security.csrf.header_name")
  def csrf_header_name() -> str:
    return "X-CSRF-Token"
  
  @static_property("autoboot.web.uniform.exception.global.code")
  def exception_global_code():
    return 500
  
  @static_property("autoboot.web.uniform.exception.global.message")
  def exception_global_message():
    return "Server error: {}"
  
  @static_property("autoboot.web.uniform.exception.request_validation.code")
  def exception_request_validation_code():
    return 400
  
  @static_property("autoboot.web.uniform.exception.request_validation.message")
  def exception_request_validation_message():
    return "Request valid error: {}"



from autoboot import AutoBoot
from autoboot.annotation import component, conditional
from nacos import NacosClient

from .nacos_properties import NacosProperties

@conditional(value="autoboot.web.discovery.nacos.enable", having_value=True)
@component()
def nacos_client():
  client = NacosClient(
    server_addresses=NacosProperties.server_address(),
    namespace=NacosProperties.namespace(),
    username=NacosProperties.username(),
    password=NacosProperties.password(),
    ak=NacosProperties.ak(),
    sk=NacosProperties.sk()
  )
  success = client.add_naming_instance(
    service_name=NacosProperties.service_name(),
    ip=NacosProperties.service_ip(),
    port=NacosProperties.service_port(),
    group_name=NacosProperties.group_name(),
    heartbeat_interval=NacosProperties.heartbeat_interval()
  )
  if success:
    AutoBoot.logger.info(f"nacos register service[{NacosProperties.service_name()}] on {NacosProperties.service_ip()}:{NacosProperties.service_port()}")
  else:
    AutoBoot.logger.error(f"nacos register service[{NacosProperties.service_name()}] fail.")
  return client



from autoboot.annotation import static_property
from nacos import client

class NacosProperties(object):
    
  @static_property("autoboot.web.discovery.nacos.enable")
  def enable():
    return False

  @static_property("autoboot.web.discovery.nacos.server_address")
  def server_address():
    return "127.0.0.1:8848"
  
  @static_property("autoboot.web.discovery.nacos.namespace")
  def namespace() -> str:
    pass
  
  @static_property("autoboot.web.discovery.nacos.username")
  def username() -> str:
    pass
  
  @static_property("autoboot.web.discovery.nacos.password")
  def password() -> str:
    pass
  
  @static_property("autoboot.web.discovery.nacos.ak")
  def ak() -> str:
    pass
  
  @static_property("autoboot.web.discovery.nacos.sk")
  def sk() -> str:
    pass

  @static_property("autoboot.web.discovery.nacos.service_name")
  def service_name():
    return "test-service"
  
  @static_property("autoboot.web.discovery.nacos.service_ip")
  def service_ip():
    return "127.0.0.1"
  
  @static_property("autoboot.web.discovery.nacos.service_port")
  def service_port():
    return 8080
  
  @static_property("autoboot.web.discovery.nacos.group_name")
  def group_name():
    return client.DEFAULT_GROUP_NAME
  
  @static_property("autoboot.web.discovery.nacos.heartbeat_interval")
  def heartbeat_interval():
    return 2



import asyncio
from typing import Optional
from importlib import import_module
from contextlib import asynccontextmanager

from autoboot import AutoBoot
from autoboot.plugin import AppPlugin
from fastapi import FastAPI, Request, status
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from starlette.middleware.sessions import SessionMiddleware

from .http_properties import HttpProperties
from .web_properties import WebProperties


@asynccontextmanager
async def app_lifespan(app: FastAPI):
  # init lifespan
  WebRunner._loop = asyncio.get_running_loop()
  AutoBoot.logger.info("hold event loop: {}.", WebRunner._loop)
  yield
  #clean up lifespan


class WebRunner(AppPlugin[FastAPI]):
  
  _loop: asyncio.AbstractEventLoop
  
  # custom context name, default is class name
  #context_name = "WebRunner"
  
  def __init__(self, scan_controllers: Optional[str | list[str]] = None) -> None:
    self._scan_controllers = scan_controllers
    
  def __repr__(self) -> str:
    return f"WebRunner(scan_controllers={self._scan_controllers}"
    
  def get_event_loop(cls) -> asyncio.AbstractEventLoop:
    return cls._loop
  
  def install(self) -> FastAPI:
    return FastAPI(lifespan=app_lifespan)
  
  def env_prepared(self) -> None:
    self.packages: list[str] = []
    if self._scan_controllers:
      self.packages.extend([self._scan_controllers] if isinstance(self._scan_controllers, str) else self._scan_controllers)
    
    if WebProperties.scan_controller_packages():
      self.packages.extend(WebProperties.scan_controller_packages()) 
      
  
  def app_started(self) -> None:
    for package in self.packages:
      import_module(package)
      
    app = self.get_context()
    
    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(request: Request, exc: RequestValidationError):
      return JSONResponse(
        status_code=status.HTTP_200_OK,
        content={
          "code": WebProperties.exception_request_validation_code(),
          "message": WebProperties.exception_request_validation_message().replace("{}", str(exc.errors()))
        },
      )
      
    # handle global exception
    @app.exception_handler(Exception)
    async def global_exception_handler(request: Request, exc: Exception):
      return JSONResponse(
          status_code=status.HTTP_200_OK,
          content={
            "code": WebProperties.exception_global_code(),
            "message": WebProperties.exception_global_message().replace("{}", str(exc))
          },
      )

    if HttpProperties.gzip_enable():
      AutoBoot.logger.info("GZIP is enabled.")
      app.add_middleware(GZipMiddleware, minimum_size=HttpProperties.gzip_minimum_size())
      
    if HttpProperties.session_enable():
      AutoBoot.logger.info("Session is enabled.")
      app.add_middleware(
        SessionMiddleware, 
        secret_key=HttpProperties.session_secret_key(),
        session_cookie=HttpProperties.session_cookie_name(),
        max_age=HttpProperties.session_max_age(),
        same_site=HttpProperties.session_same_site(),
        https_only=HttpProperties.session_https_only()
      )
    
    if WebProperties.cross_origin():
      AutoBoot.logger.info("Cross-Origin Resource Sharing (CORS) is enabled.")
      from fastapi.middleware.cors import CORSMiddleware
      app.add_middleware(
        CORSMiddleware,
        allow_origins=WebProperties.allow_origins(),
        allow_credentials=WebProperties.allow_credentials(),
        allow_methods=WebProperties.allow_methods(),
        allow_headers=WebProperties.allow_headers(),
        max_age=WebProperties.max_age()
      )
      
    if WebProperties.csrf_enable():
      AutoBoot.logger.info("CSRF is enabled.")
      try:
        from autoboot_web.middleware import UniformCSRFMiddleware
      except ImportError:
        raise ImportError("CSRF middleware is not installed.")
    
      app.add_middleware(
        UniformCSRFMiddleware, 
        secret=WebProperties.csrf_secret(), 
        cookie_name=WebProperties.csrf_cookie_name(),
        cookie_domain=WebProperties.csrf_cookie_domain(),
        header_name=WebProperties.csrf_header_name()
      )



import uuid
from autoboot.annotation import static_property

class HttpProperties:
  
  @static_property("autoboot.web.http.gzip.enable")
  def gzip_enable() -> bool:
    return False

  @static_property("autoboot.web.http.gzip.minimum_size")
  def gzip_minimum_size() -> int:
    return 1000

  @static_property("autoboot.web.http.session.enable")
  def session_enable() -> bool:
    return False
  
  @static_property("autoboot.web.http.session.secret")
  def session_secret_key() -> str:
    return str(uuid.uuid1())

  @static_property("autoboot.web.http.session.cookie_name")
  def session_cookie_name() -> str:
    return "session_id"
  
  @static_property("autoboot.web.http.session.max_age")
  def session_max_age() -> int:
    return 3600

  @static_property("autoboot.web.http.session.same_site")
  def session_same_site() -> str:
    return "lax"
  
  @static_property("autoboot.web.http.session.https_only")
  def session_https_only() -> bool:
    return False
  
  
  



from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette_csrf import CSRFMiddleware


class UniformCSRFMiddleware(CSRFMiddleware):
  
  def _get_error_response(self, request: Request) -> Response:
    return JSONResponse(content={'code': 400, 'message': 'CSRF Failed'}, status_code=200)



from typing import Type, Optional
from fastapi.routing import APIRouter
from autoboot_web import WebRunner
from autoboot_web.mvc import ClassBasedView


def Controller(path: Optional[str] = None, tag: Optional[str] = None):
  """
  Controller decorator

  :param path: path to the controller
  :param tag: tag for the api doc
  """
  
  if path:
    if not path.startswith("/"):
      path = "/" + path
  
  def wrapper(cls) -> Type[ClassBasedView]:
    router = APIRouter(tags=[tag] if tag else None)
    
    items = cls.__dict__.items()
    for name, method in items:
      if not callable(method) or not hasattr(method, "req_method"):
        continue
      
      if not method.__path__:
        raise Exception("Must set path on request mapping")
      else:
        if path:
          if path.endswith("/") and method.__path__.startswith("/"):
            method.__path__ = path + method.__path__[1:]
          elif not path.endswith("/") and not method.__path__.startswith("/"):
            method.__path__ = path + "/" + method.__path__
          else:
            method.__path__ = path + method.__path__

        if not method.__path__.startswith("/"):
          method.__path__ = "/" + method.__path__
          
        if method.req_method == "GET":
          router.add_api_route(
              path=method.__path__,
              endpoint=method,
              methods=["GET"],
              **method.__kwargs__
          )
        elif method.req_method == "POST":
          router.add_api_route(
              path=method.__path__,
              endpoint=method,
              methods=["POST"],
              **method.__kwargs__
          )
        elif method.req_method == "PUT":
          router.add_api_route(
              path=method.__path__,
              endpoint=method,
              methods=["PUT"],
              **method.__kwargs__
          )
        elif method.req_method == "DELETE":
          router.add_api_route(
              path=method.__path__,
              endpoint=method,
              methods=["DELETE"],
              **method.__kwargs__
          )
        elif method.req_method == "PATCH":
          router.add_api_route(
              path=method.__path__,
              endpoint=method,
              methods=["PATCH"],
              **method.__kwargs__
          )
        else:
          raise Exception("Unknown request method")
        
    # Add get_router method to the class
    cls.get_router = classmethod(lambda: router)
    
    # rectify routes which bind on method
    classBasedView = ClassBasedView(router=router, cls=cls)
    
    # register router
    WebRunner.get_context().include_router(router)
    
    return classBasedView
  return wrapper


def Get(path: str, **kwargs):
  def decorator(fn):
    fn.req_method = "GET"
    fn.__path__ = path
    fn.__kwargs__ = kwargs
    return fn
  return decorator

def Post(path: str, **kwargs):
  def decorator(func):
      func.req_method = "POST"
      func.__path__ = path
      func.__kwargs__ = kwargs
      return func
  return decorator

def Delete(path: str, **kwargs):
  def decorator(func):
      func.req_method = "DELETE"
      func.__path__ = path
      func.__kwargs__ = kwargs
      return func
  return decorator

def Put(path: str, **kwargs):
  def decorator(func):
      func.req_method = "PUT"
      func.__path__ = path
      func.__kwargs__ = kwargs
      return func
  return decorator

def Patch(path: str, **kwargs):
  def decorator(func):
      func.req_method = "PATCH"
      func.__path__ = path
      func.__kwargs__ = kwargs
      return func
  return decorator